from scrapy import Request
from scrapy_splash import SplashRequest


class RequestFactory:
    def __init__(self, javascript):
        self.javascript = javascript

    def create(self, url, next_page):
        if self.javascript:
            return SplashRequest(url, args={"wait": 0.5}, meta={"page": next_page})
        else:
            return Request(url, meta={"page": next_page})



import pytest

from webcomix.exceptions import NextLinkNotFound
from webcomix.scrapy.crawler_worker import CrawlerWorker
from webcomix.scrapy.verification.verification_spider import VerificationSpider
from webcomix.tests.fake_websites.fixture import one_webpage_uri


def test_spider_raising_error_gets_raised_by_crawler_worker(one_webpage_uri):
    settings = {"LOG_ENABLED": False}
    worker = CrawlerWorker(
        settings,
        False,
        VerificationSpider,
        start_url=one_webpage_uri,
        next_page_selector="//div/@href",
        comic_image_selector="//img/@src",
        number_of_pages_to_check=2,
    )

    with pytest.raises(NextLinkNotFound):
        worker.start()



class CrawlerBlocked(Exception):
    pass


class NextLinkNotFound(Exception):
    def __init__(self, failed_url, next_page_xpath):
        self.failed_url = failed_url
        self.next_page_xpath = next_page_xpath



import click
from click.testing import CliRunner
import pytest

from webcomix import cli
from webcomix.comic import Comic
from webcomix.exceptions import CrawlerBlocked, NextLinkNotFound
from webcomix.supported_comics import supported_comics
from webcomix.tests.fake_websites.fixture import (
    three_webpages_uri,
    three_webpages_alt_text_uri,
)

first_comic = list(sorted(supported_comics.keys()))[0]


def test_print_verification(capfd, three_webpages_uri):
    comic = Comic("test", three_webpages_uri, "//img/@src", "//a/@href")
    verification = comic.verify_xpath()
    cli.print_verification(verification)
    out, err = capfd.readouterr()

    three_webpages_folder = three_webpages_uri.strip("1.html")

    assert out == (
        "Page 1:\n"
        "Page URL: " + three_webpages_uri + "\n"
        "Image URLs:\n"
        "" + three_webpages_folder + "1.jpeg"
        "\n"
        "\n"
        "Page 2:\n"
        "Page URL: " + three_webpages_folder + "2.html"
        "\n"
        "Image URLs:\n"
        "" + three_webpages_folder + "2.jpeg"
        "\n"
        "\n"
        "Page 3:\n"
        "Page URL: " + three_webpages_folder + "3.html"
        "\n"
        "Image URLs:\n"
        "\n"
        "\n"
    )


def test_print_verification_with_alt_text(capfd, three_webpages_alt_text_uri):
    comic = Comic(
        "test_alt",
        three_webpages_alt_text_uri,
        "//img/@src",
        "//a/@href",
        alt_text="//img/@title",
    )
    verification = comic.verify_xpath()
    cli.print_verification(verification)
    out, err = capfd.readouterr()

    three_webpages_alt_text_folder = three_webpages_alt_text_uri.strip("1.html")

    assert out == (
        "Page 1:\n"
        "Page URL: " + three_webpages_alt_text_uri + "\n"
        "Image URLs:\n"
        "" + three_webpages_alt_text_folder + "1.jpeg"
        "\n"
        "Alt text: First page\n"
        "\n"
        "Page 2:\n"
        "Page URL: " + three_webpages_alt_text_folder + "2.html"
        "\n"
        "Image URLs:\n"
        "" + three_webpages_alt_text_folder + "2.jpeg"
        "\n"
        "Alt text: Second page\n"
        "\n"
        "Page 3:\n"
        "Page URL: " + three_webpages_alt_text_folder + "3.html"
        "\n"
        "Image URLs:\n"
        "\n"
        "\n"
    )


def test_print_verification_with_no_validation_throws_crawler_blocked():
    with pytest.raises(CrawlerBlocked):
        cli.print_verification(None)


def test_comics():
    runner = CliRunner()
    result = runner.invoke(cli.comics)
    assert result.exit_code == 0
    assert len(result.output) > 0


def test_good_download_ends_up_downloading(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")

    result = runner.invoke(cli.download, [first_comic])
    assert result.exit_code == 0
    assert mock_download.call_count == 1


def test_predefined_unknown_comic_does_not_download(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")

    result = runner.invoke(cli.download, ["foo"])
    assert result.exit_code == 2
    assert mock_download.call_count == 0


def test_predefined_downloadable_comic_downloads_the_comic(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")

    result = runner.invoke(cli.download, [first_comic])
    assert result.exit_code == 0
    assert mock_download.call_count == 1


def test_predefined_downloadable_comic_makes_the_cbz_file(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_convert_to_cbz = mocker.patch("webcomix.comic.Comic.convert_to_cbz")

    result = runner.invoke(cli.download, [first_comic, "--cbz"])
    assert result.exit_code == 0
    assert mock_convert_to_cbz.call_count == 1


def test_predefined_unknown_comic_does_not_make_the_cbz_file(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_convert_to_cbz = mocker.patch("webcomix.comic.Comic.convert_to_cbz")

    result = runner.invoke(cli.download, ["foo", "--cbz"])
    assert result.exit_code == 2
    assert mock_convert_to_cbz.call_count == 0


def test_custom_comic_asks_for_verification_before_downloading(mocker):
    runner = CliRunner()
    mock_manager = mocker.Mock()
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_verify_xpath = mocker.patch("webcomix.comic.Comic.verify_xpath")
    mock_print_verification = mocker.patch("webcomix.cli.print_verification")
    mock_manager.attach_mock(mock_download, "download")
    mock_manager.attach_mock(mock_verify_xpath, "verify_xpath")
    mock_manager.attach_mock(mock_print_verification, "print_verification")

    result = runner.invoke(
        cli.custom,
        [
            "foo",
            "--start_url=url",
            "--next_page_xpath=next_page",
            "--image_xpath=image",
        ],
        "yes",
    )
    assert result.exit_code == 0
    mock_manager.assert_has_calls(
        [
            mocker.call.verify_xpath(),
            mocker.call.print_verification(mocker.ANY),
            mocker.call.download(),
        ]
    )


def test_custom_comic_makes_the_cbz_file(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_verify_xpath = mocker.patch("webcomix.comic.Comic.verify_xpath")
    mock_print_verification = mocker.patch("webcomix.cli.print_verification")
    mock_convert_to_cbz = mocker.patch("webcomix.comic.Comic.convert_to_cbz")

    result = runner.invoke(
        cli.custom,
        [
            "foo",
            "--start_url=url",
            "--next_page_xpath=next_page",
            "--image_xpath=image",
            "--cbz",
        ],
        "y",
    )
    assert result.exit_code == 0
    assert mock_convert_to_cbz.call_count == 1


def test_custom_comic_doesnt_ask_for_verification_if_next_link_not_found(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_verify_xpath = mocker.patch(
        "webcomix.comic.Comic.verify_xpath",
        side_effect=NextLinkNotFound(mocker.ANY, mocker.ANY),
    )
    mock_print_verification = mocker.patch("webcomix.cli.print_verification")

    result = runner.invoke(
        cli.custom,
        [
            "foo",
            "--start_url=url",
            "--next_page_xpath=next_page",
            "--image_xpath=image",
        ],
        "yes",
    )

    assert result.exit_code == 1
    assert type(result.exception) is SystemExit
    assert mock_verify_xpath.call_count == 1
    assert mock_print_verification.call_count == 0
    assert mock_download.call_count == 0


def test_custom_comic_doesnt_download_comic_if_crawler_blocked(mocker):
    runner = CliRunner()
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_verify_xpath = mocker.patch("webcomix.comic.Comic.verify_xpath")
    mock_print_verification = mocker.patch(
        "webcomix.cli.print_verification", side_effect=CrawlerBlocked()
    )

    result = runner.invoke(
        cli.custom,
        [
            "foo",
            "--start_url=url",
            "--next_page_xpath=next_page",
            "--image_xpath=image",
        ],
        "yes",
    )

    assert result.exit_code == 1
    assert type(result.exception) is SystemExit
    assert mock_verify_xpath.call_count == 1
    assert mock_print_verification.call_count == 1
    assert mock_download.call_count == 0


def test_discovered_comic_searches_for_a_comic(mocker):
    runner = CliRunner()
    mock_discovery = mocker.patch(
        "webcomix.cli.discovery",
        return_value=(
            Comic(mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY),
            mocker.ANY,
        ),
    )
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_verify_xpath = mocker.patch("webcomix.comic.Comic.verify_xpath")
    mock_print_verification = mocker.patch("webcomix.cli.print_verification")

    result = runner.invoke(cli.search, ["foo", "--start_url=good"], "y")
    assert result.exit_code == 0
    assert mock_discovery.call_count == 1


def test_discovered_comic_asks_for_verification_before_downloading(mocker):
    runner = CliRunner()
    mock_manager = mocker.Mock()
    mock_discovery = mocker.patch(
        "webcomix.cli.discovery",
        return_value=(
            Comic(mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY),
            mocker.ANY,
        ),
    )
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_verify_xpath = mocker.patch("webcomix.comic.Comic.verify_xpath")
    mock_print_verification = mocker.patch("webcomix.cli.print_verification")
    mock_manager.attach_mock(mock_download, "download")
    mock_manager.attach_mock(mock_verify_xpath, "verify_xpath")
    mock_manager.attach_mock(mock_print_verification, "print_verification")

    result = runner.invoke(cli.search, ["foo", "--start_url=good"], "y")
    assert result.exit_code == 0
    mock_manager.assert_has_calls(
        [mocker.call.print_verification(mocker.ANY), mocker.call.download()]
    )


def test_discovered_comic_makes_cbz_file(mocker):
    runner = CliRunner()
    mock_discovery = mocker.patch(
        "webcomix.cli.discovery",
        return_value=(
            Comic(mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY),
            mocker.ANY,
        ),
    )
    mock_download = mocker.patch("webcomix.comic.Comic.download")
    mock_verify_xpath = mocker.patch("webcomix.comic.Comic.verify_xpath")
    mock_print_verification = mocker.patch("webcomix.cli.print_verification")
    mock_convert_to_cbz = mocker.patch("webcomix.comic.Comic.convert_to_cbz")

    result = runner.invoke(cli.search, ["foo", "--start_url=good", "--cbz"], "y")
    assert result.exit_code == 0
    assert mock_convert_to_cbz.call_count == 1



import os
from typing import List, Mapping, Dict, Any
from urllib.parse import urlparse
from zipfile import ZipFile, BadZipFile

import click

from webcomix.scrapy.download.comic_spider import ComicSpider
from webcomix.scrapy.verification.verification_spider import VerificationSpider
from webcomix.scrapy.crawler_worker import CrawlerWorker
from webcomix.scrapy.verification.web_page import WebPage

SPLASH_SETTINGS = {
    "SPLASH_URL": "http://0.0.0.0:8050",
    "DOWNLOADER_MIDDLEWARES": {
        "scrapy_splash.SplashCookiesMiddleware": 723,
        "scrapy_splash.SplashMiddleware": 725,
        "scrapy.downloadermiddlewares.httpcompression.HttpCompressionMiddleware": 810,
    },
    "SPIDER_MIDDLEWARES": {"scrapy_splash.SplashDeduplicateArgsMiddleware": 100},
    "DUPEFILTER_CLASS": "scrapy_splash.SplashAwareDupeFilter",
    "HTTPCACHE_STORAGE": "scrapy_splash.SplashAwareFSCacheStorage",
}

FAKE_USERAGENT_SETTINGS = {
    "DOWNLOADER_MIDDLEWARES": {
        "scrapy.downloadermiddlewares.useragent.UserAgentMiddleware": None,
        "scrapy.downloadermiddlewares.retry.RetryMiddleware": None,
        "scrapy_fake_useragent.middleware.RandomUserAgentMiddleware": 400,
        "scrapy_fake_useragent.middleware.RetryUserAgentMiddleware": 401,
    },
    "FAKEUSERAGENT_PROVIDERS": [
        "scrapy_fake_useragent.providers.FakeUserAgentProvider",
        "scrapy_fake_useragent.providers.FakerProvider",
        "scrapy_fake_useragent.providers.FixedUserAgentProvider",
    ],
    "FAKEUSERAGENT_FALLBACK": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36",
}


class Comic:
    def __init__(
        self,
        name: str,
        start_url: str,
        comic_image_selector: str,
        next_page_selector: str,
        start_page: int = 1,
        alt_text: str = None,
        single_page: bool = False,
        javascript: bool = False,
        title: bool = False,
        debug: bool = False,
    ):
        self.name = name
        self.start_url = start_url
        self.start_page = start_page
        self.next_page_selector = next_page_selector
        self.comic_image_selector = comic_image_selector
        self.alt_text = alt_text
        self.single_page = single_page
        self.javascript = javascript
        self.title = title
        self.debug = debug

    def download(self) -> None:
        """
        Downloads an entire comic page by page starting from the first one
        and saves them in the directory_name created in the current working
        directory
        """
        if not os.path.isdir(self.name):
            os.makedirs(self.name)

        settings = {
            **FAKE_USERAGENT_SETTINGS,
            "ITEM_PIPELINES": {
                "webcomix.scrapy.download.comic_pipeline.ComicPipeline": 1,
                "scrapy.pipelines.files.FilesPipeline": 500,
            },
            "LOG_ENABLED": self.debug,
            "FILES_STORE": self.name,
            "MEDIA_ALLOW_REDIRECTS": True,
            "RETRY_HTTP_CODES": [500, 502, 503, 504, 522, 524, 403, 408, 429],
        }  # type: Dict

        if self.javascript:
            settings.update(SPLASH_SETTINGS)

        worker = CrawlerWorker(
            settings,
            False,
            ComicSpider,
            start_url=self.start_url,
            start_page=self.start_page,
            comic_image_selector=self.comic_image_selector,
            next_page_selector=self.next_page_selector,
            directory=self.name,
            javascript=self.javascript,
            title=self.title,
            alt_text=self.alt_text,
        )

        worker.start()

        click.echo("Finished downloading the images.")

    def convert_to_cbz(self) -> None:
        """
        Takes all of the previously downloaded pages and compresses them in
        a .cbz file, erasing them afterwards.
        """
        with ZipFile("{}.cbz".format(self.name), mode="a") as cbz_file:
            images = os.listdir(self.name)
            for image in images:
                image_location = "{}/{}".format(self.name, image)
                cbz_file.write(image_location, image)
                os.remove(image_location)
            os.rmdir(self.name)
            if cbz_file.testzip() is not None:
                raise BadZipFile(
                    "Error while testing the archive; it might be corrupted."
                )

    def verify_xpath(self) -> List[Mapping[str, WebPage]]:
        """
        Takes a url and the XPath expressions for the next_page and image to
        go three pages into the comic. It returns a tuple containing the url
        of each page and their respective image urls.
        """
        settings = {
            **FAKE_USERAGENT_SETTINGS,
            "LOG_ENABLED": self.debug,
        }  # type: Dict[str, Any]

        if self.javascript:
            settings.update(SPLASH_SETTINGS)

        worker = CrawlerWorker(
            settings,
            True,
            VerificationSpider,
            start_url=self.start_url,
            comic_image_selector=self.comic_image_selector,
            next_page_selector=self.next_page_selector,
            number_of_pages_to_check=1 if self.single_page else 3,
            javascript=self.javascript,
            alt_text=self.alt_text,
        )

        verification = worker.start()

        return verification

    @staticmethod
    def save_image_location(
        url: str, page: int, directory_name: str = "", title: bool = False
    ) -> str:
        """
        Returns the relative location in the filesystem under which the
        webcomic will be saved. If directory_name is specified, it will be
        relative to the current directory; if not specified, it will return
        the name relative to the directory in which it is downloaded.
        """
        file_name = Comic.save_image_filename(url, page, title, directory_name)
        return os.path.join(directory_name, file_name)

    @staticmethod
    def save_image_filename(
        url: str, page: int, title_present: bool = False, comic_name: str = ""
    ) -> str:
        """
        Returns the filename of the comic image depending on whether or not we
        want it the comic's name to be present.
        """
        if url.count(".") <= 1:
            # No file extension (only dot in url is domain name)
            return str(page)

        parsed_filepath = urlparse(url).path
        file_extension = parsed_filepath[parsed_filepath.rindex(".") :]
        if title_present:
            return "{}-{}{}".format(comic_name, page, file_extension)
        else:
            return "{}{}".format(page, file_extension)

    @staticmethod
    def save_alt_text_location(page: int, directory_name: str = "") -> str:
        """
        Returns the relative location in the filesystem under which the comic
        image's alt text will be saved
        """
        file_name = str.format("{}.txt", page)
        return os.path.join(directory_name, file_name)



from webcomix.comic import Comic
from webcomix.search import discovery
from webcomix.tests.fake_websites.fixture import (
    one_webpage_searchable_uri,
    three_webpages_uri,
    three_webpages_classes_uri,
)


def test_search_searchable_website(mocker, three_webpages_classes_uri):
    expected = Comic(
        "Blindsprings",
        three_webpages_classes_uri,
        "//*[contains(translate(@class, 'ABCDEFGHIJKLMNOPQRSTUVWXYZ','abcdefghijklmnopqrstuvwxyz'), 'comic')]//@src",
        "//*[contains(translate(@class, 'ABCDEFGHIJKLMNOPQRSTUVWXYZ','abcdefghijklmnopqrstuvwxyz'), 'next')]//@href",
    )
    mocker.patch("webcomix.search.possible_image_xpath", ["comic"])
    mocker.patch("webcomix.search.possible_next_page_xpath", ["next"])
    mocker.patch("webcomix.search.possible_tags_image", ["*"])
    mocker.patch("webcomix.search.possible_tags_next", ["*"])
    mocker.patch("webcomix.search.possible_attributes_image", ["@class"])
    mocker.patch("webcomix.search.possible_attributes_next", ["@class"])
    mocker.patch("webcomix.util.check_first_pages")
    comic, result = discovery("Blindsprings", three_webpages_classes_uri)

    three_webpages_classes_folder = three_webpages_classes_uri.strip("1.html")

    assert result == [
        {
            "page": 1,
            "url": three_webpages_classes_uri,
            "image_urls": [three_webpages_classes_folder + "1.jpeg"],
            "alt_text": None,
        },
        {
            "page": 2,
            "url": three_webpages_classes_folder + "2.html",
            "image_urls": [three_webpages_classes_folder + "2.jpeg"],
            "alt_text": None,
        },
        {
            "page": 3,
            "url": three_webpages_classes_folder + "3.html",
            "image_urls": [three_webpages_classes_folder + "3.jpeg"],
            "alt_text": None,
        },
    ]

    assert comic.start_url == expected.start_url
    assert comic.next_page_selector == expected.next_page_selector
    assert comic.comic_image_selector == expected.comic_image_selector


def test_search_unsearchable_website(mocker, three_webpages_uri):
    mocker.patch("webcomix.search.possible_image_xpath", ["comic"])
    mocker.patch("webcomix.search.possible_next_page_xpath", ["next"])
    mocker.patch("webcomix.search.possible_tags_image", ["*"])
    mocker.patch("webcomix.search.possible_tags_next", ["*"])
    mocker.patch("webcomix.search.possible_attributes_image", ["@class"])
    mocker.patch("webcomix.search.possible_attributes_next", ["@class"])

    assert discovery("test", three_webpages_uri) == (None, None)


def test_can_stop_searching(mocker, three_webpages_classes_uri):
    mocker.patch("webcomix.search.possible_image_xpath", ["comic"])
    mocker.patch("webcomix.search.possible_next_page_xpath", ["next"])
    mocker.patch("webcomix.search.possible_tags_image", ["div"])
    mocker.patch("webcomix.search.possible_tags_next", ["div"])
    mocker.patch("webcomix.search.possible_attributes_image", ["@rel"])
    mocker.patch("webcomix.search.possible_attributes_next", ["@class"])
    exit_called = mocker.patch("sys.exit")
    mocker.patch("webcomix.comic.Comic.verify_xpath", side_effect=KeyboardInterrupt)
    result = discovery("test", three_webpages_classes_uri)
    assert exit_called.call_count == 1
    assert result == (None, None)


def test_can_find_single_page_correctly_while_searching(
    mocker, one_webpage_searchable_uri
):
    mocker.patch("webcomix.search.possible_image_xpath", ["image"])
    mocker.patch("webcomix.search.possible_next_page_xpath", ["next"])
    mocker.patch("webcomix.search.possible_tags_image", ["*"])
    mocker.patch("webcomix.search.possible_tags_next", ["*"])
    mocker.patch("webcomix.search.possible_attributes_image", ["@class"])
    mocker.patch("webcomix.search.possible_attributes_next", ["."])

    comic, result = discovery("test", one_webpage_searchable_uri, single_page=True)

    validation = comic.verify_xpath()

    assert len(result) == 1
    assert result == validation
    assert len(result[0]["image_urls"]) == 2



import pytest

from webcomix.comic import Comic
from webcomix.supported_comics import supported_comics
from webcomix.util import check_first_pages


@pytest.mark.slow
@pytest.mark.parametrize("comic_name", list(supported_comics.keys()))
def test_supported_comics(comic_name):
    comic = Comic(comic_name, *supported_comics[comic_name])
    first_pages = comic.verify_xpath()
    check_first_pages(first_pages)



import os

from scrapy.exceptions import DropItem
import pytest

from webcomix.scrapy.download.comic_pipeline import ComicPipeline
from webcomix.scrapy.download.comic_page import ComicPage
from webcomix.supported_comics import supported_comics

first_comic = list(sorted(supported_comics.values()))[0]

expected_url_image = "http://imgs.xkcd.com/comics/barrel_cropped_(1).jpg"
expected_image_location = "test/1.jpg"
expected_image_filename = "1.jpg"


def test_get_media_requests_returns_good_request_when_file_not_present(mocker):
    mocker.patch("os.path.isfile", return_value=False)
    mock_spider_info = mocker.patch("scrapy.pipelines.media.MediaPipeline.SpiderInfo")
    mocker.patch(
        "webcomix.comic.Comic.save_image_location", return_value=expected_image_location
    )
    mocker.patch(
        "webcomix.comic.Comic.save_image_filename", return_value=expected_image_filename
    )
    pipeline = ComicPipeline(store_uri="foo")
    elements = list(
        pipeline.get_media_requests(
            ComicPage(url=expected_url_image, page=1, title=False, alt_text=None),
            mock_spider_info,
        )
    )
    request = elements[0]
    assert request.url == expected_url_image
    assert request.meta["image_file_name"] == expected_image_filename
    os.rmdir("foo")


def test_get_media_requests_drops_item_when_file_present(mocker):
    mocker.patch("os.path.isfile", return_value=True)
    mock_spider_info = mocker.patch("scrapy.pipelines.media.MediaPipeline.SpiderInfo")
    mocker.patch(
        "webcomix.comic.Comic.save_image_location", return_value=expected_image_location
    )
    pipeline = ComicPipeline(store_uri="foo")
    with pytest.raises(DropItem):
        list(
            pipeline.get_media_requests(
                ComicPage(url=expected_url_image, page=1, title=False, alt_text=None),
                mock_spider_info,
            )
        )
    os.rmdir("foo")


def test_get_media_requests_drops_item_when_file_present_in_zip(mocker):
    mocker.patch("os.path.isfile", side_effect=[False])
    mocker.patch(
        "webcomix.scrapy.download.comic_pipeline.ComicPipeline.image_in_zipfile",
        return_value=True,
    )
    mock_spider_info = mocker.patch("scrapy.pipelines.media.MediaPipeline.SpiderInfo")
    mocker.patch(
        "webcomix.comic.Comic.save_image_location", return_value=expected_image_location
    )
    pipeline = ComicPipeline(store_uri="foo")
    with pytest.raises(DropItem):
        list(
            pipeline.get_media_requests(
                ComicPage(url=expected_url_image, page=1, title=False, alt_text=None),
                mock_spider_info,
            )
        )
    os.rmdir("foo")


def test_item_completed_returns_item_when_file_downloaded(mocker):
    results = [(True, {"path": expected_image_location})]
    item = ComicPage()
    pipeline = ComicPipeline(store_uri="foo")

    result = pipeline.item_completed(results, item, mocker.ANY)

    assert result == item
    os.rmdir("foo")


def test_item_completed_returns_drops_when_file_not_downloaded(mocker):
    results = [(False, {})]
    item = ComicPage()
    pipeline = ComicPipeline(store_uri="foo")

    with pytest.raises(DropItem):
        pipeline.item_completed(results, item, mocker.ANY)
    os.rmdir("foo")


def test_file_path_is_image_path(mocker):
    mock_request = mocker.patch("scrapy.http.Request")
    mock_request.meta = {"image_file_name": expected_image_location}
    pipeline = ComicPipeline(store_uri="foo")
    file_path = pipeline.file_path(mock_request)
    assert file_path == expected_image_location
    os.rmdir("foo")


def test_image_not_in_zip_if_zip_does_not_exist(mocker):
    mocker.patch("os.path.isfile", return_value=False)
    assert ComicPipeline.image_in_zipfile(mocker.ANY, mocker.ANY) == False



import sys
from itertools import product
from typing import Optional, List, Tuple, Mapping

import click
from tqdm import tqdm

from webcomix.comic import Comic
from webcomix.util import check_first_pages

possible_next_page_xpath = ["next"]
possible_image_xpath = ["comic", "image"]
possible_tags_image = ["*", "img", "div"]
possible_tags_next = ["*", "a", "div", "li"]
possible_attributes_image = [".", "@src", "@class", "@id", "@alt"]
possible_attributes_next = [".", "text()", "@class", "@id", "@alt", "@rel"]


def discovery(
    name: str,
    url: str,
    start_page: int = 1,
    alt_text: str = None,
    single_page: bool = False,
    javascript: bool = False,
    title: bool = False,
    debug: bool = False,
) -> Tuple[Optional[Comic], Optional[List[Mapping]]]:
    def to_lower_case(attribute):
        return (
            "translate({}, "
            "'ABCDEFGHIJKLMNOPQRSTUVWXYZ',"
            "'abcdefghijklmnopqrstuvwxyz')"
        ).format(attribute)

    click.echo("Looking for a path to the whole comic... (Ctrl-C to exit)")
    combinations = product(
        possible_next_page_xpath,
        possible_image_xpath,
        possible_tags_image,
        possible_tags_next,
        possible_attributes_image,
        possible_attributes_next,
    )
    total = (
        len(possible_next_page_xpath)
        * len(possible_image_xpath)
        * len(possible_tags_image)
        * len(possible_tags_next)
        * len(possible_attributes_image)
        * len(possible_attributes_next)
    )

    for next_page, image, tag_image, tag_next, attribute_image, attribute_next in tqdm(
        combinations, total=total
    ):
        next_page_xpath = "//{}[contains({}, '{}')]//@href".format(
            tag_next, to_lower_case(attribute_next), next_page
        )
        image_xpath = "//{}[contains({}, '{}')]//@src".format(
            tag_image, to_lower_case(attribute_image), image
        )
        try:
            comic = Comic(
                name,
                url,
                image_xpath,
                next_page_xpath,
                start_page=start_page,
                alt_text=alt_text,
                single_page=single_page,
                javascript=javascript,
                title=title,
                debug=debug,
            )
            first_pages = comic.verify_xpath()
            check_first_pages(first_pages)
            return comic, first_pages
        except KeyboardInterrupt:
            sys.exit(0)
        except:
            continue
    click.echo("Search has failed.")
    return None, None



from operator import itemgetter
import os
from zipfile import ZipFile

import click
import scrapy
from scrapy.exceptions import DropItem
from scrapy.pipelines.files import FilesPipeline

from webcomix.comic import Comic


class ComicPipeline(FilesPipeline):
    def get_media_requests(self, item, info):
        click.echo("Saving image {}".format(item.get("url")))
        url, page, title, alt_text = itemgetter("url", "page", "title", "alt_text")(
            item
        )
        image_path_directory = Comic.save_image_location(
            url, page, info.spider.directory, title
        )
        if os.path.isfile(image_path_directory) or self.image_in_zipfile(
            item, info.spider.directory
        ):
            click.echo("The image was already downloaded. Skipping...")
            raise DropItem("The image was already downloaded. Skipping...")
        if alt_text is not None:
            with open(
                Comic.save_alt_text_location(page, info.spider.directory), "w"
            ) as alt_text_file:
                alt_text_file.write(alt_text)
        yield scrapy.Request(
            item.get("url"),
            meta={
                "image_file_name": Comic.save_image_filename(
                    item.get("url"), item.get("page"), title, info.spider.directory
                )
            },
        )

    def item_completed(self, results, item, info):
        file_paths = [data["path"] for ok, data in results if ok]
        if not file_paths:
            click.echo("Could not find comic image.")
            raise DropItem("Could not find comic image.")

        return item

    def file_path(self, request, response=None, info=None, *, item=None):
        return request.meta.get("image_file_name")

    @staticmethod
    def image_in_zipfile(item, directory):
        zipfile_path = "{}.cbz".format(directory)
        if not os.path.isfile(zipfile_path):
            return False
        image_path_cbz = Comic.save_image_location(item.get("url"), item.get("page"))
        with ZipFile(zipfile_path, "r") as zipfile:
            return image_path_cbz in zipfile.namelist()



def check_first_pages(first_pages):
    page_links = set([page.get("url") for page in first_pages])
    assert len(set(page_links)) == len(page_links)
    assert len(page_links) == len(first_pages)
    list_of_images = [page.get("image_urls") for page in first_pages]
    image_links = [image for page in list_of_images for image in page]
    assert len(set(image_links)) == len(image_links)
    assert len(image_links) >= len(first_pages)



import os
from pathlib import Path

import pytest


def get_dir_path_of_script():
    return Path(os.path.dirname(os.path.abspath(__file__)))


@pytest.fixture
def three_webpages_uri():
    return str(
        get_dir_path_of_script().joinpath("three_webpages/1.html").resolve().as_uri()
    )


@pytest.fixture
def three_webpages_alt_text_uri():
    return str(
        get_dir_path_of_script()
        .joinpath("three_webpages_alt_text/1.html")
        .resolve()
        .as_uri()
    )


@pytest.fixture
def three_webpages_classes_uri():
    return str(
        get_dir_path_of_script()
        .joinpath("three_webpages_classes/1.html")
        .resolve()
        .as_uri()
    )


@pytest.fixture
def one_webpage_uri():
    return str(
        get_dir_path_of_script().joinpath("one_webpage/1.html").resolve().as_uri()
    )


@pytest.fixture
def one_webpage_searchable_uri():
    return str(
        get_dir_path_of_script()
        .joinpath("one_webpage_searchable/1.html")
        .resolve()
        .as_uri()
    )



import docker
import pytest

from webcomix.docker import DockerManager, CONTAINER_NAME


@pytest.fixture
def cleanup_container():
    yield None
    client = docker.from_env()
    for container in client.containers.list():
        if container.attrs["Config"]["Image"] == CONTAINER_NAME:
            container.kill()


def test_no_javascript_spawns_no_container(cleanup_container):
    manager = DockerManager(False)
    manager.__enter__()
    manager.client = docker.from_env()
    assert manager._get_container() is None


def test_javascript_spawns_container(cleanup_container):
    manager = DockerManager(True)
    manager.__enter__()
    assert manager._get_container() is not None
    manager.__exit__(None, None, None)


def test_javascript_exit_removes_container(cleanup_container):
    manager = DockerManager(True)
    manager.__enter__()
    manager.__exit__(None, None, None)
    assert manager._get_container() is None



import signal
from multiprocessing import Process, Queue

from pydispatch import dispatcher
from scrapy import signals
from scrapy.crawler import CrawlerProcess


class CrawlerWorker(Process):
    def __init__(self, settings, return_items, *crawl_args, **crawl_kwargs):
        super().__init__(daemon=True)
        self.result_queue = Queue()
        self.crawl_args = crawl_args
        self.crawl_kwargs = crawl_kwargs

        self.process = CrawlerProcess(settings)
        self.kill_process = False
        dispatcher.connect(self._spider_error, signals.spider_error)
        signal.signal(signal.SIGINT, self._exit_gracefully)
        signal.signal(signal.SIGTERM, self._exit_gracefully)

    def _spider_error(self, failure):
        self.result_queue.put(failure.value)

    def _exit_gracefully(self, signum, frame):
        self.kill_process = True
        self.process.stop()

    def run(self):
        self.process.crawl(
            *self.crawl_args, **self.crawl_kwargs, result_queue=self.result_queue
        )
        self.process.start()

    def start(self):
        super().start()
        super().join()

        result = []
        while not self.result_queue.empty():
            result.append(self.result_queue.get())

        if self.kill_process:
            raise KeyboardInterrupt
        if len(result) == 1 and isinstance(result[0], Exception):
            raise result[0]
        elif not result:
            return None
        else:
            return result



from urllib.parse import urljoin

from scrapy import Spider

from webcomix.exceptions import CrawlerBlocked, NextLinkNotFound
from webcomix.scrapy.request_factory import RequestFactory
from webcomix.scrapy.util import is_not_end_of_comic
from webcomix.scrapy.verification.web_page import WebPage


class VerificationSpider(Spider):
    name = "Verification Spider"

    def __init__(self, *args, **kwargs):
        self.start_url = kwargs.get("start_url")
        self.next_page_selector = kwargs.get("next_page_selector", None)
        self.comic_image_selector = kwargs.get("comic_image_selector", None)
        self.number_of_pages_to_check = kwargs.get("number_of_pages_to_check", 3)
        javascript = kwargs.get("javascript", False)
        self.alt_text_selector = kwargs.get("alt_text")
        self.result_queue = kwargs.get("result_queue")
        self.request_factory = RequestFactory(javascript)
        super(VerificationSpider, self).__init__(*args, **kwargs)

    def start_requests(self):
        yield self.request_factory.create(url=self.start_url, next_page=1)

    def parse(self, response):
        comic_image_urls = response.xpath(self.comic_image_selector).getall()
        page = response.meta.get("page") or 1
        image_urls = [
            urljoin(response.url, image_element_url.strip())
            for image_element_url in comic_image_urls
        ]
        alt_text = (
            response.xpath(self.alt_text_selector).get()
            if self.alt_text_selector is not None
            else None
        )
        next_page_url = response.xpath(self.next_page_selector).get()
        if page >= self.number_of_pages_to_check:
            self.result_queue.put(
                WebPage(
                    url=response.url,
                    page=page,
                    image_urls=image_urls,
                    alt_text=alt_text,
                )
            )
        elif is_not_end_of_comic(next_page_url):
            self.result_queue.put(
                WebPage(
                    url=response.url,
                    page=page,
                    image_urls=image_urls,
                    alt_text=alt_text,
                )
            )
            yield self.request_factory.create(
                url=response.urljoin(next_page_url).strip(), next_page=page + 1
            )
        else:
            self.result_queue.put(
                NextLinkNotFound(response.url, self.next_page_selector)
            )



from scrapy import Request
from scrapy_splash import SplashRequest

from webcomix.scrapy.request_factory import RequestFactory


AN_URL = "https://xkcd.com"


def test_factory_creates_request_without_javascript(mocker):
    request_factory = RequestFactory(False)
    request = request_factory.create(AN_URL, mocker.ANY)
    assert type(request) == Request


def test_factory_creates_splash_request_with_javascript(mocker):
    request_factory = RequestFactory(True)
    request = request_factory.create(AN_URL, mocker.ANY)
    assert type(request) == SplashRequest



from urllib.parse import urljoin

import click
from scrapy import Spider

from webcomix.scrapy.download.comic_page import ComicPage
from webcomix.exceptions import CrawlerBlocked
from webcomix.scrapy.request_factory import RequestFactory
from webcomix.scrapy.util import is_not_end_of_comic


class ComicSpider(Spider):
    name = "Comic Spider"
    handle_httpstatus_list = [403]

    def __init__(self, *args, **kwargs):
        self.start_url = kwargs.get("start_url")
        self.next_page_selector = kwargs.get("next_page_selector", None)
        self.comic_image_selector = kwargs.get("comic_image_selector", None)
        self.start_page = kwargs.get("start_page", 1)
        self.directory = kwargs.get("directory", None)
        javascript = kwargs.get("javascript", False)
        self.alt_text = kwargs.get("alt_text", None)
        self.title = kwargs.get("title", False)
        self.result_queue = kwargs.get("result_queue")
        self.request_factory = RequestFactory(javascript)
        super(ComicSpider, self).__init__(*args, **kwargs)

    def start_requests(self):
        yield self.request_factory.create(url=self.start_url, next_page=self.start_page)

    def parse(self, response):
        click.echo("Downloading page {}".format(response.url))
        comic_image_urls = response.xpath(self.comic_image_selector).getall()
        page = response.meta.get("page") or self.start_page
        alt_text = (
            response.xpath(self.alt_text).get() if self.alt_text is not None else None
        )
        for index, comic_image_url in enumerate(comic_image_urls):
            yield ComicPage(
                url=urljoin(response.url, comic_image_url.strip()),
                page=page + index,
                title=self.title,
                alt_text=alt_text,
            )
        if not comic_image_urls:
            click.echo("Could not find comic image.")
        next_page_url = response.xpath(self.next_page_selector).get()
        if is_not_end_of_comic(next_page_url):
            yield self.request_factory.create(
                url=response.urljoin(next_page_url).strip(),
                next_page=page + len(comic_image_urls),
            )



import docker
from time import sleep

CONTAINER_NAME = "scrapinghub/splash"


class DockerManager:
    def __init__(self, javascript):
        self.javascript = javascript

    def __enter__(self):
        if self.javascript:
            self.client = docker.from_env()
            running = self._container_running()
            if running:
                return None
            self.client.containers.run(
                CONTAINER_NAME, detach=True, ports={"8050/tcp": 8050}
            )
            print("Container started")
            sleep(0.5)
            print("Container ready")

    def __exit__(self, exc_type, exc_value, tb):
        if self.javascript:
            container = self._get_container()
            if container is not None:
                container.kill()
        if exc_type is not None:
            return False
        return True

    def _container_running(self):
        return self._get_container() is not None

    def _get_container(self):
        containers = self.client.containers.list()
        for container in containers:
            if container.attrs["Config"]["Image"] == CONTAINER_NAME:
                return container
        return None



supported_comics = {
    "xkcd": (
        "https://xkcd.com/1/",
        "//div[@id='comic']//img/@src",
        "//a[@rel='next']/@href",
    ),
    "xkcd_alt": (
        "https://xkcd.com/1/",
        "//div[@id='comic']//img/@src",
        "//a[@rel='next']/@href",
        "//div[@id='comic']//img/@title",
    ),
    "Nedroid": (
        "http://nedroid.com/2005/09/2210-whee/",
        "//div[@id='comic']/img/@src",
        "//div[@class='nav-next']/a/@href",
    ),
    "JL8": (
        "https://limbero.org/jl8/1",
        "//img[@alt='Comic']/@src",
        "//a[text()='>']/@href",
    ),
    "SMBC": (
        "https://www.smbc-comics.com/comic/2002-09-05",
        "//img[@id='cc-comic']/@src",
        "//a[@class='cc-next']/@href",
    ),
    "Blindsprings": (
        "https://www.blindsprings.com/comic/blindsprings-cover-book-one",
        "//img[@id='cc-comic']/@src",
        "//a[@class='cc-next']/@href",
    ),
    "TheAbominableCharlesChristopher": (
        "https://karlkerschl.com/comic/episode-one/",
        "//div[@id='comic']//img/@src",
        "//a[contains(@class, 'comic-nav-next')]/@href",
    ),
    "GuildedAge": (
        "https://guildedage.net/comic/chapter-1-cover/",
        "//div[@id='comic']//img/@src",
        "//a[@class='navi comic-nav-next navi-next']/@href",
    ),
    "TalesOfElysium": (
        "https://ssp-comics.com/comics/toe/?page=1&mode=10",
        "//div[@id='ImageComicContainer']//img[contains(@src, 'comic')]/@src",
        "//a[button/@id='next10Button']/@href",
    ),
    "AmazingSuperPowers": (
        "https://www.amazingsuperpowers.com/2007/09/heredity/",
        "//div[@class='comicpane']/img/@src",
        "//a[@class='navi navi-next']/@href",
    ),
    "Gunshow": (
        "https://gunshowcomic.com/1",
        "//img[@class='strip']/@src",
        "//span[@class='snavb'][4]/a/@href",
    ),
    "Lackadaisy": (
        "https://www.lackadaisycats.com/comic.php?comicid=1",
        "//div[@id='content']/img/@src",
        "//div[@class='next']/a/@href",
    ),
    "WildeLife": (
        "https://www.wildelifecomic.com/comic/1",
        "//img[@id='cc-comic']/@src",
        "//a[@class='cc-next']/@href",
    ),
    "ElGoonishShive": (
        "https://www.egscomics.com/comic/2002-01-21",
        "//div[@id='cc-comicbody']//img/@src",
        "//a[@class='cc-next']/@href",
    ),
    "StandStillStaySilent": (
        "https://www.sssscomic.com/comic.php?page=1",
        "//img[@class='comicnormal' and contains(@src, 'comicpages')]/@src",
        "//a[img[contains(@src, 'next')]]/@href",
    ),
    "SchlockMercenary": (
        "https://www.schlockmercenary.com/2000-06-12",
        "//div[@class='strip-images']//img/@src",
        "//a[@class='next-strip']/@href",
    ),
    "Nerfnow": (
        "https://www.nerfnow.com/comic/4",
        "//div[@id='comic']//img/@src",
        "//li[@id='nav_next']//a/@href",
    ),
    "EdisonRex": (
        "https://www.edisonrex.net/comic/issue-1-cover",
        "//div[@id='cc-comicbody']//img/@src",
        "//a[@class='cc-next']/@href",
    ),
    "Spinnerette": (
        "https://www.spinnyverse.com/comic/02-09-2010",
        "//div[@id='cc-comicbody']//img/@src",
        "//a[@class='cc-next']/@href",
    ),
    "RyanMadeMistakes": (
        "http://mistakes.ryanestrada.com/comic/0001",
        "//img[@id='cc-comic']/@src",
        "//a[@class='cc-next']/@href",
    ),
    "MissingMonday": (
        "https://www.missingmondaycomic.com/comic/chapter-01-page-01",
        "//img[@id='cc-comic']/@src",
        "//a[@class='cc-next']/@href",
    ),
    "StarTrip": (
        "https://www.startripcomic.com/comic/chapter-1-cover",
        "//img[@id='cc-comic']/@src",
        "//a[@class='cc-next']/@href",
    ),
    "StupidFox": (
        "http://stupidfox.net/hello",
        "//img[contains(@src, 'stupidfox.net/art/')]/@src",
        "//span[@class='spriteNext']/parent::a/@href",
    ),
    "GunnerkriggCourt": (
        "https://www.gunnerkrigg.com/?p=1",
        "//img[@class='comic_image']/@src",
        "//a[img[contains(@src, 'next')]]/@href",
    ),
    "Housepets": (
        "https://www.housepetscomic.com/comic/2008/06/02/when-boredom-strikes/",
        "//div[@id='comic']//img/@src",
        "//div[@id='comic']//@href",
    ),
}



import os
import shutil
from zipfile import ZipFile, BadZipFile

import pytest

from webcomix.comic import Comic, SPLASH_SETTINGS
from webcomix.supported_comics import supported_comics
from webcomix.tests.fake_websites.fixture import (
    three_webpages_uri,
    three_webpages_alt_text_uri,
    one_webpage_uri,
)


@pytest.fixture
def cleanup_test_directories():
    if os.path.isfile("xkcd.cbz"):
        os.remove("xkcd.cbz")
    if os.path.isfile("test.cbz"):
        os.remove("test.cbz")
    if os.path.isdir("xkcd"):
        shutil.rmtree("xkcd")
    if os.path.isdir("test"):
        shutil.rmtree("test")
    yield None
    if os.path.isdir("xkcd"):
        shutil.rmtree("xkcd")
    if os.path.isdir("test"):
        shutil.rmtree("test")
    if os.path.isfile("xkcd.cbz"):
        os.remove("xkcd.cbz")
    if os.path.isfile("test.cbz"):
        os.remove("test.cbz")


@pytest.fixture
def fake_downloaded_xkcd_comic():
    if os.path.isfile("xkcd.cbz"):
        os.remove("xkcd.cbz")
    if os.path.isdir("xkcd"):
        shutil.rmtree("xkcd")
    comic = Comic(
        "xkcd",
        "http://xkcd.com/1/",
        "//div[@id='comic']/img/@src",
        "//a[@rel='next']/@href",
        False,
    )
    os.mkdir("xkcd")
    for i in range(1, 6):
        with open("xkcd/{}.txt".format(i), "w") as image_file:
            image_file.write("testing {}".format(i))
    yield comic
    if os.path.isfile("xkcd.cbz"):
        os.remove("xkcd.cbz")
    if os.path.isdir("xkcd"):
        shutil.rmtree("xkcd")


def test_save_image_location():
    assert (
        Comic.save_image_location(
            "http://imgs.xkcd.com/comics/barrel_cropped_(1).jpg", 1, "foo"
        )
        == "foo/1.jpg"
    )
    assert Comic.save_image_location("", 1, "bar") == "bar/1"
    assert (
        Comic.save_image_location(
            "http://imgs.xkcd.com/comics/barrel_cropped_(1).jpg?q=123", 1, "foo"
        )
        == "foo/1.jpg"
    )


def test_save_image_filename_with_title_present():
    assert (
        Comic.save_image_filename(
            "http://imgs.xkcd.com/comics/barrel_cropped_(1).jpg", 1, True, "foo"
        )
        == "foo-1.jpg"
    )


def test_make_cbz(fake_downloaded_xkcd_comic):
    fake_downloaded_xkcd_comic.convert_to_cbz()
    with ZipFile("xkcd.cbz") as cbz_file:
        for i in range(1, 6):
            with cbz_file.open("{}.txt".format(i), "r") as image_file:
                assert image_file.read().decode("ascii") == "testing {}".format(i)


def test_make_cbz_corrupted_archive(mocker, capfd, fake_downloaded_xkcd_comic):
    mocker.patch.object(ZipFile, "testzip", return_value=mocker.ANY)
    with pytest.raises(BadZipFile):
        fake_downloaded_xkcd_comic.convert_to_cbz()


def test_download_runs_the_worker(mocker, cleanup_test_directories):
    mock_crawler_running = mocker.patch(
        "webcomix.scrapy.crawler_worker.CrawlerWorker.start"
    )
    comic = Comic(
        "xkcd",
        "http://xkcd.com/1/",
        "//div[@id='comic']//img/@src",
        "//a[@rel='next']/@href",
        False,
    )
    comic.download()
    assert mock_crawler_running.call_count == 1


def test_download_saves_the_files(cleanup_test_directories, three_webpages_uri):
    comic = Comic("test", three_webpages_uri, "//img/@src", "//a/@href")
    comic.download()
    path, dirs, files = next(os.walk("test"))
    assert len(files) == 2


def test_download_saves_the_files_with_correct_first_item(
    cleanup_test_directories, three_webpages_uri
):
    comic = Comic("test", three_webpages_uri, "//img/@src", "//a/@href", start_page=10)
    comic.download()
    files = next(os.walk("test"))[2]
    assert sorted(files) == ["10", "11"]


def test_download_with_alt_text_saves_the_text(
    cleanup_test_directories, three_webpages_alt_text_uri
):
    comic = Comic(
        "test",
        three_webpages_alt_text_uri,
        "//img/@src",
        "//a/@href",
        alt_text="//img/@title",
    )
    comic.download()
    path, dirs, files = next(os.walk("test"))
    assert len(files) == 4


def test_download_does_not_add_crawlers_in_main_process(
    mocker, cleanup_test_directories, three_webpages_uri
):
    mocker.patch("webcomix.scrapy.crawler_worker.CrawlerWorker.start")
    mock_add_to_crawl = mocker.patch("scrapy.crawler.Crawler.crawl")
    comic = Comic("test", three_webpages_uri, "//img/@src", "//a/@href")
    comic.download()
    assert mock_add_to_crawl.call_count == 0


def test_convert_to_cbz_adds_all_files_to_cbz(
    cleanup_test_directories, three_webpages_uri
):
    comic = Comic("test", three_webpages_uri, "//img/@src", "//a/@href")
    comic.download()
    comic.convert_to_cbz()
    with ZipFile("{}.cbz".format(comic.name), mode="r") as cbz_file:
        assert len(cbz_file.infolist()) == 2


def test_verify_xpath(three_webpages_uri):
    comic = Comic("test", three_webpages_uri, "//img/@src", "//a/@href")

    three_webpages_folder = three_webpages_uri.strip("1.html")

    assert comic.verify_xpath() == [
        {
            "page": 1,
            "url": three_webpages_uri,
            "image_urls": [three_webpages_folder + "1.jpeg"],
            "alt_text": None,
        },
        {
            "page": 2,
            "url": three_webpages_folder + "2.html",
            "image_urls": [three_webpages_folder + "2.jpeg"],
            "alt_text": None,
        },
        {
            "page": 3,
            "url": three_webpages_folder + "3.html",
            "image_urls": [],
            "alt_text": None,
        },
    ]


def test_verify_xpath_with_alt_text(three_webpages_alt_text_uri):
    comic = Comic(
        "test_alt",
        three_webpages_alt_text_uri,
        "//img/@src",
        "//a/@href",
        alt_text="//img/@title",
    )

    three_webpages_alt_text_folder = three_webpages_alt_text_uri.strip("1.html")

    assert comic.verify_xpath() == [
        {
            "page": 1,
            "url": three_webpages_alt_text_uri,
            "image_urls": [three_webpages_alt_text_folder + "1.jpeg"],
            "alt_text": "First page",
        },
        {
            "page": 2,
            "url": three_webpages_alt_text_folder + "2.html",
            "image_urls": [three_webpages_alt_text_folder + "2.jpeg"],
            "alt_text": "Second page",
        },
        {
            "page": 3,
            "url": three_webpages_alt_text_folder + "3.html",
            "image_urls": [],
            "alt_text": None,
        },
    ]


def test_verify_xpath_only_verifies_one_page_with_single_page(one_webpage_uri):
    comic = Comic("test", one_webpage_uri, "//img/@src", "//a/@href", single_page=True)
    actual = comic.verify_xpath()
    assert len(actual) == 1
    assert len(actual[0]["image_urls"]) == 2


def test_download_will_run_splash_settings_if_javascript(mocker):
    mocker.patch("os.path.isdir")
    mock_crawler_worker = mocker.patch("webcomix.comic.CrawlerWorker")
    comic = Comic(
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        True,
    )
    comic.download()
    settings = mock_crawler_worker.call_args_list[0][0][0]
    assert all(setting in settings.items() for setting in SPLASH_SETTINGS.items())


def test_download_will_not_run_splash_settings_if_not_javascript(mocker):
    mocker.patch("os.path.isdir")
    mock_crawler_worker = mocker.patch("webcomix.comic.CrawlerWorker")
    comic = Comic(mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY)
    comic.download()
    settings = mock_crawler_worker.call_args_list[0][0][0]
    assert all(setting not in settings.items() for setting in SPLASH_SETTINGS.items())


def test_verify_xpath_will_run_splash_settings_if_javascript(mocker):
    mocker.patch("os.path.isdir")
    mock_crawler_worker = mocker.patch("webcomix.comic.CrawlerWorker")
    comic = Comic(
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        True,
    )
    comic.verify_xpath()
    settings = mock_crawler_worker.call_args_list[0][0][0]
    assert all(setting in settings.items() for setting in SPLASH_SETTINGS.items())


def test_verify_xpath_will_not_run_splash_settings_if_not_javascript(mocker):
    mocker.patch("os.path.isdir")
    mock_crawler_worker = mocker.patch("webcomix.comic.CrawlerWorker")
    comic = Comic(mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY, mocker.ANY, False)
    comic.verify_xpath()
    settings = mock_crawler_worker.call_args_list[0][0][0]
    assert all(setting not in settings.items() for setting in SPLASH_SETTINGS.items())



#! python3

import click

from webcomix.comic import Comic
from webcomix.exceptions import CrawlerBlocked, NextLinkNotFound
from webcomix.search import discovery
from webcomix.supported_comics import supported_comics
from webcomix.docker import DockerManager


@click.group()
@click.version_option()
def cli():
    pass


@cli.command()
def comics():
    """
    Shows all predefined comics
    """
    comics_content = [
        "{}: {}".format(key, value[0])
        for key, value in sorted(supported_comics.items())
    ]

    click.echo("\n".join(comics_content))


@cli.command()
@click.argument(
    "name", type=click.Choice([k for k, v in sorted(supported_comics.items())])
)
@click.option(
    "--cbz", is_flag=True, default=False, help="Outputs the comic as a cbz file"
)
@click.option(
    "--title", is_flag=True, default=False, help="Add title of comic in image names"
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Add debugging output"
)
def download(name, cbz, title, verbose):
    """
    Downloads a predefined comic by name
    """
    if name in list(supported_comics.keys()):
        comic = Comic(name, *supported_comics[name], title=title, debug=verbose)
        download_webcomic(comic, cbz)


@cli.command()
@click.argument("name", type=click.STRING)
@click.option(
    "--start-url",
    "--start_url",
    prompt=True,
    type=click.STRING,
    help="URL of the comic's first page",
)
@click.option(
    "--start-page",
    "--start_page",
    type=click.INT,
    default=1,
    help="Number of comic's first page to be downloaded",
)
@click.option(
    "--cbz", default=False, is_flag=True, help="Outputs the comic as a cbz file"
)
@click.option(
    "--single-page",
    "--single_page",
    default=False,
    is_flag=True,
    help="Downloads from a single webpage",
)
@click.option(
    "--javascript",
    "--js",
    "-j",
    default=False,
    is_flag=True,
    help="Renders javascript in the page (slower)",
)
@click.option(
    "--title", is_flag=True, default=False, help="Add title of comic in image names"
)
@click.option(
    "--alt-text",
    "-a",
    default=None,
    type=click.STRING,
    help="Optional XPath to fetch an additionnal text while scraping",
)
@click.option(
    "--yes", "-y", default=False, is_flag=True, help="Skips the verification prompt"
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Add debugging output"
)
def search(
    name,
    start_url,
    start_page,
    cbz,
    single_page,
    javascript,
    title,
    alt_text,
    yes,
    verbose,
):
    """
    Downloads a webcomic using a general XPath
    """
    with DockerManager(javascript):
        comic, validation = discovery(
            name,
            start_url,
            start_page,
            alt_text,
            single_page,
            javascript,
            title,
            verbose,
        )
        if comic is not None:
            print_verification(validation)
            click.echo("Verify that the links above are correct.")
            if yes or click.confirm("Are you sure you want to proceed?"):
                download_webcomic(comic, cbz)


@cli.command()
@click.argument("name", type=click.STRING)
@click.option(
    "--start-url",
    "--start_url",
    prompt=True,
    type=click.STRING,
    help="URL of the comic's first page to be downloaded",
)
@click.option(
    "--start-page",
    "--start_page",
    type=click.INT,
    default=1,
    help="Number of comic's first page to be downloaded",
)
@click.option(
    "--image-xpath",
    "--image_xpath",
    prompt=True,
    type=click.STRING,
    help="XPath expression giving the url to the image",
)
@click.option(
    "--next-page-xpath",
    "--next_page_xpath",
    prompt=True,
    type=click.STRING,
    help="XPath expression giving the url to the next page",
)
@click.option(
    "--cbz", default=False, is_flag=True, help="Outputs the comic as a cbz file"
)
@click.option(
    "--single-page",
    "--single_page",
    "-s",
    default=False,
    is_flag=True,
    help="Downloads from a single webpage",
)
@click.option(
    "--javascript",
    "--js",
    "-j",
    default=False,
    is_flag=True,
    help="Renders javascript in the page (slower)",
)
@click.option(
    "--title", is_flag=True, default=False, help="Add title of comic in image names"
)
@click.option(
    "--alt-text",
    "-a",
    default=None,
    type=click.STRING,
    help="XPath to fetch an additionnal text while scraping",
)
@click.option(
    "--yes", "-y", default=False, is_flag=True, help="Skips the verification prompt"
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Add debugging output"
)
def custom(
    name,
    start_url,
    start_page,
    next_page_xpath,
    image_xpath,
    cbz,
    single_page,
    javascript,
    title,
    alt_text,
    yes,
    verbose,
):
    """
    Downloads a user-defined webcomic
    """
    with DockerManager(javascript):
        comic = Comic(
            name,
            start_url,
            image_xpath,
            next_page_xpath,
            start_page,
            alt_text,
            single_page,
            javascript,
            title,
            verbose,
        )
        try:
            validation = comic.verify_xpath()
        except NextLinkNotFound as exception:
            click.echo(
                "Could not find next link of: {} \nwith next page XPath expression: {}".format(
                    exception.failed_url, exception.next_page_xpath
                )
            )
            click.echo(
                "Have you tried testing your XPath expression with 'scrapy shell'?"
            )
            raise click.Abort()
        try:
            print_verification(validation)
        except CrawlerBlocked as exception:
            click.echo("{} could not be accessed with webcomix.".format(name))
            click.echo(
                "Chances are the website you're trying to download images from doesn't want to be scraped."
            )
            raise click.Abort()
        click.echo("Verify that the links above are correct.")
        if yes or click.confirm("Are you sure you want to proceed?"):
            download_webcomic(comic, cbz)


def print_verification(validation):
    """
    Prints the verification given by the verify_xpath function
    """
    if validation is None:
        raise CrawlerBlocked()
    for item in sorted(validation, key=lambda x: x.get("page")):
        output = "Page {}:\nPage URL: {}\nImage URLs:\n{}\n".format(
            item.get("page"), item.get("url"), "\n".join(item.get("image_urls"))
        )
        if item.get("alt_text") is not None:
            output += "Alt text: {}\n".format(item.get("alt_text"))
        click.echo(output)


def download_webcomic(comic, cbz):
    comic.download()
    if cbz:
        comic.convert_to_cbz()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['webcomix',
 'webcomix.scrapy',
 'webcomix.scrapy.download',
 'webcomix.scrapy.download.tests',
 'webcomix.scrapy.tests',
 'webcomix.scrapy.verification',
 'webcomix.tests',
 'webcomix.tests.fake_websites']

package_data = \
{'': ['*'],
 'webcomix.tests.fake_websites': ['one_webpage/*',
                                  'one_webpage_searchable/*',
                                  'three_webpages/*',
                                  'three_webpages_alt_text/*',
                                  'three_webpages_classes/*']}

install_requires = \
['Scrapy>=2.5.1,<3.0.0',
 'click>=7.1.1,<8.0.0',
 'docker>=5.0.3,<6.0.0',
 'scrapy-fake-useragent>=1.4.4,<2.0.0',
 'scrapy-splash>=0.8.0,<0.9.0',
 'tqdm>=4.43.0,<5.0.0']

entry_points = \
{'console_scripts': ['webcomix = webcomix.cli:cli']}

setup_kwargs = {
    'name': 'webcomix',
    'version': '3.6.1',
    'description': 'Webcomic downloader',
    'long_description': '# webcomix\n\n[![Build Status](https://github.com/J-CPelletier/webcomix/workflows/Build/badge.svg?branch=master)](https://github.com/J-CPelletier/webcomix/actions)[![Coverage Status](https://coveralls.io/repos/github/J-CPelletier/webcomix/badge.svg?branch=master)](https://coveralls.io/github/J-CPelletier/webcomix?branch=master)[![PyPI version](https://badge.fury.io/py/webcomix.svg)](https://badge.fury.io/py/webcomix)\n\n## Description\n\nwebcomix is a webcomic downloader that can additionally create a .cbz (Comic Book ZIP) file once downloaded.\n\n## Notice\n\nThis program is for personal use only. Please be aware that by making the downloaded comics publicly available without the permission of the author, you may be infringing upon various copyrights.\n\n## Installation\n\n### Dependencies\n\n* Python (3.6 or newer)\n* click\n* scrapy (Some additional steps might be required to include this package and can be found [here](https://doc.scrapy.org/en/latest/intro/install.html#intro-install-platform-notes))\n* scrapy-splash\n* scrapy-fake-useragent\n* tqdm\n* [Docker](https://www.docker.com/) (To be able to download JavaScript-dependent websites with `-j` option)\n\n### Process\n\n#### End user\n\n1. Install [Python 3](https://www.python.org/downloads/)\n2. Install the command line interface tool with `pip install webcomix`\n\n#### Developer\n\n1. Install [Python 3](https://www.python.org/downloads/)\n2. Clone this repository and open a terminal in its directory\n3. Install [poetry](https://github.com/python-poetry/poetry) with `pip install poetry`\n3. Download the dependencies by running `poetry install`\n4. Install pre-commit hooks with `pre-commit install`\n\n## Usage\n\n`webcomix [OPTIONS] COMMAND [ARGS]`\n\n### Global Flags\n\n#### help\n\nShow the help message and exit.\n\n#### Version\n\nShow the version number and exit.\n\n### Commands\n\n#### comics\n\nShows all predefined comics which can be used with the `download` command.\n\n#### download\n\nDownloads a predefined comic. Supports the `--cbz` flag, which creates a .cbz archive of the downloaded comic.\n\n#### search\n\nSearches for an XPath that can download the whole comic. Supports the `--cbz` flag, which creates a .cbz archive of the downloaded comic,`-s`, which verifies only the provided page of the comic, `-y`, which skips the verification prompt, and `-j`, which runs the javascript on pages before downloading.\n\n#### custom\n\nDownloads a user-defined comic. To download a specific comic, you\'ll need a link to the first page, an XPath expression giving out the link to the next page and an XPath expression giving out the link to the image. More info [here](http://www.w3schools.com/xml/xpath_syntax.asp). Supports the `--cbz` flag, which creates a .cbz archive of the downloaded comic, `-s`, which verifies only the provided page of the comic, and `-y`, which skips the verification prompt.\n\n### Examples\n\n* `webcomix download xkcd`\n* `webcomix search xkcd --start-url=http://xkcd.com/1/`\n* `webcomix custom --cbz` (You will be prompted about other needed arguments)\n* `webcomix custom xkcd --start-url=http://xkcd.com/1/ --next-page-xpath="//a[@rel=\'next\']/@href" --image-xpath="//div[@id=\'comic\']//img/@src" --cbz` (Same as before, but with all arguments declared beforehand)\n\n### Making an XPath selector\n\nUsing an HTML inspector, spot a html path to the next link\'s `href` attribute/comic image\'s `src` attribute.\n\ne.g.: `//div[@class=\'foo\']/img/@src`\nThis will select the src attribute of the first image whose class is: foo\n\nNote: `webcomix` works best on static websites, since `scrapy`(the framework we use to travel web pages) doesn\'t process Javascript.\n\nTo make sure your XPath is correct, you have to go into `scrapy shell`, which should be downloaded when you\'ve installed `webcomix`.\n\n```\nscrapy shell <website> --> Use the website\'s url to go to it.\n> response.body --> Will give you the html from the website.\n> response.xpath --> Test an xpath selection. If you get [], this means your XPath expression hasn\'t gotten anything from the webpage.\n```\n\n## Contribution\n\nThe procedure depends on the type of contribution:\n\n* If you simply want to request the addition of a comic to the list of supported comics, make an issue with the label "Enhancement".\n* If you want to request the addition of a feature to the system or a bug fix, make an issue with the appropriate label.\n\n### Running the tests\n\nTo run the tests, you have to use the `pytest` command in the webcomix folder.\n',
    'author': 'Jean-Christophe Pelletier',
    'author_email': 'pelletierj97@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/J-CPelletier/webcomix',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



from multiprocessing import Queue

import pytest

from webcomix.scrapy.download.comic_spider import ComicSpider
from webcomix.exceptions import CrawlerBlocked


def test_parse_yields_good_page(mocker):
    mock_response = mocker.patch("scrapy.http.Response")
    mock_response.urljoin.return_value = "http://xkcd.com/3/"
    mock_response.url = "http://xkcd.com/2/"
    mock_selector = mocker.patch("scrapy.selector.SelectorList")
    mock_response.xpath.return_value = mock_selector
    mock_selector.getall.return_value = ["//imgs.xkcd.com/comics/tree_cropped_(1).jpg"]
    mock_selector.get.return_value = "xkcd.com/3/"

    spider = ComicSpider()
    result = spider.parse(mock_response)
    results = list(result)
    assert len(results) == 2
    assert results[0].get("url") == "http://imgs.xkcd.com/comics/tree_cropped_(1).jpg"
    assert results[1].url == "http://xkcd.com/3/"


def test_parse_yields_multiple_subpages(mocker):
    mock_response = mocker.patch("scrapy.http.Response")
    mock_response.urljoin.return_value = "http://xkcd.com/3/"
    mock_response.url = "http://xkcd.com/2/"
    mock_selector = mocker.patch("scrapy.selector.SelectorList")
    mock_response.xpath.return_value = mock_selector
    mock_selector.getall.return_value = [
        "//imgs.xkcd.com/comics/tree_cropped_(1).jpg",
        "//imgs.xkcd.com/comics/tree_cropped_(2).jpg",
    ]
    mock_selector.get.return_value = "xkcd.com/3/"

    spider = ComicSpider()
    result = spider.parse(mock_response)
    results = list(result)
    assert len(results) == 3
    assert results[0].get("url") == "http://imgs.xkcd.com/comics/tree_cropped_(1).jpg"
    assert results[1].get("url") == "http://imgs.xkcd.com/comics/tree_cropped_(2).jpg"
    assert results[2].url == "http://xkcd.com/3/"


def test_parse_strips_additionnal_spaces(mocker):
    mock_response = mocker.patch("scrapy.http.Response")
    mock_response.urljoin.return_value = " http://xkcd.com/3/ "
    mock_response.url = "http://xkcd.com/2/"
    mock_selector = mocker.patch("scrapy.selector.SelectorList")
    mock_response.xpath.return_value = mock_selector
    mock_selector.getall.return_value = [
        " //imgs.xkcd.com/comics/tree_cropped_(1).jpg ",
        " //imgs.xkcd.com/comics/tree_cropped_(2).jpg ",
    ]
    mock_selector.get.return_value = " xkcd.com/3/ "

    spider = ComicSpider()
    result = spider.parse(mock_response)
    results = list(result)
    assert len(results) == 3
    assert results[0].get("url") == "http://imgs.xkcd.com/comics/tree_cropped_(1).jpg"
    assert results[1].get("url") == "http://imgs.xkcd.com/comics/tree_cropped_(2).jpg"
    assert results[2].url == "http://xkcd.com/3/"


def test_parse_yields_no_pages(mocker):
    mock_response = mocker.patch("scrapy.http.Response")
    mock_response.urljoin.return_value = "http://xkcd.com/3/"
    mock_response.url = "http://xkcd.com/2/"
    mock_selector = mocker.patch("scrapy.selector.SelectorList")
    mock_response.xpath.return_value = mock_selector
    mock_selector.getall.return_value = []
    mock_selector.get.return_value = "xkcd.com/3/"

    spider = ComicSpider()
    result = spider.parse(mock_response)
    results = list(result)
    assert len(results) == 1
    assert results[0].url == "http://xkcd.com/3/"
